/* Copyright (c) 2011, 2016, Oracle America, Inc. and its Affiliates
 * All rights reserved.
 *    Author: Knut Omang <knut.omang@oracle.com>
 *
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its contributors
 *    may be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * This file is part of the user space library for
 * Oracle Scalable Infiniband Fabric (SIF) Host Channel Adapters
 *
 * sif.h: main entry points and initialization for userspace library
 */

#ifndef _SIF_H
#define _SIF_H
#if defined(__cplusplus)
extern "C" {
#endif

#include <stddef.h>

#include <infiniband/driver.h>
#include <infiniband/arch.h>

#ifndef PCI_VENDOR_ID_SUN
#define PCI_VENDOR_ID_SUN       0x108E /* pcidatabase.com */
#endif

#ifndef PCI_DEVICE_ID_PSIF_PF
#define PCI_DEVICE_ID_PSIF_PF       0x2088
#define PCI_DEVICE_ID_PSIF_VF       0x2089
#define PCI_DEVICE_ID_SIBS_PF       0x2188
#define PCI_DEVICE_ID_SIBS_VF       0x2189
#endif

#define PSIF_REVISION(sdevice) ((sdevice)->revision)

#define HIDDEN      __attribute__((visibility ("hidden")))

#define PFX     "sif: "

/* SIF_OPT_DISABLE_FLAGS: */
#define SOD_EXPLICIT_WC 0x1
#define SOD_WC_TO_SQ 0x2

enum sif_vendor_flags
{
    MMU_special   =  0x1,  /* Use special mmu setup in associated mappings
                            * NB! Modifies input to ibv_reg_mr! */
    SQ_mode       =  0x2,  /* Trigger send queue mode instead of using VCBs */
    proxy_mode    =  0x4,  /* Enable EPS-A proxying - requires the eps_a field to be set */
    SVF_kernel_mode =  0x8,  /* Enable kernel mode - default is direct user mode */
    tsu_qosl      = 0x10,  /* Value to use for the qosl bit in the qp state */
    no_checksum   = 0x20,  /* No csum for qp, wqe.wr.csum = qp.magic */
};

enum sif_mem_type
{
    SIFMT_BYPASS,    /* Use MMU bypass in associated mmu contexts */
    SIFMT_UMEM,      /* Normal default umem based user level mapping */
    SIFMT_UMEM_SPT,  /* Mapping of user memory based on the process' own page table */
    SIFMT_CS,        /* A large (sparsely populated) SIF only vaddr mapping (used for a.o.SQ CMPL) */
    SIFMT_ZERO,      /* Special mapping of a vaddr range to fewer pages (see #1931) */
    SIFMT_BYPASS_RO, /* MMU bypass mapped read only for device (requires IOMMU enabled) */
    SIFMT_UMEM_RO,   /* GVA2GPA mapped read only for device (requires IOMMU enabled) */
    SIFMT_MAX
};

enum sif_proxy_type
{
    SIFPX_OFF, /* Default value - no proxying */
    SIFPX_EPSA_1,
    SIFPX_EPSA_2,
    SIFPX_EPSA_3,
    SIFPX_EPSA_4
};

enum sif_flush_type
{
    NO_FLUSH,
    FLUSH_SQ,
    FLUSH_RQ
};

struct sif_device {
    struct verbs_device verbs_dev;
    uint16_t            abi_version;
    uint16_t            revision;
    uint32_t 		mr_size;  /* Maximal number of MRs in the system */
    uint64_t		min_resp_ms; /* Universal timeout scaling for sif */
};


enum desc_type
{
    SIFD_SQ_SW,
    SIFD_SQ_HW,
    SIFD_RQ,
    SIFD_CQ,
    SIFD_MAX
};


struct sif_table {
    uint32_t ext_sz;          /* Dist.in bytes between start of each entry */
    uint32_t entry_per_block; /* Number of entries per block of descriptors */
    uint32_t block_sz;        /* ext_sz * entry_per_block */
};

struct sif_map {
    uint16_t cmd;    /* mmap command used to create the map */
    uint32_t index;  /* index used in mapping */
    void* base;
    uint64_t sz;
};

struct sif_sq_hdl
{
    uint64_t wr_id;  /* Stored work id */
    uint32_t sq_seq; /* Extra sanity checks */
    int used;
};

struct sif_sq {
    struct sif_map m;
    uint32_t entries;
    uint32_t mask;     /* entries - 1 for modulo using & */
    uint32_t extent;
    uint32_t sge_entries;
    uint32_t sgl_offset;
    uint32_t sg_mr_idx; /* lkey for the sge list in the send queue */
    uint64_t dma_handle; /* dma address for the (bypass mode) sg_mr */
    struct sif_cq *cq;
    struct sif_sq_hdl *wr_hdl; /* map from sq entry index to wr_id + optional bookkeeping */
    uint8_t complete_all; /* Gets or'ed into completion bit in WRs */
    volatile struct sif_sq_sw *sw; /* kernel mapped sw descriptor entries */
    struct psif_sq_hw *hw; /* kernel mapped hw descriptor entries */
    pthread_spinlock_t lock; /* Serializes operations on this queue */
};

struct sif_rq {
    struct sif_map m;
    struct verbs_srq verbs_srq; /* Only used if this RQ is used a SRQ */
    uint32_t index;
    uint32_t entries;
    uint32_t mask;     /* entries - 1 for modulo using & */
    uint32_t extent;
    uint32_t sge_entries;
    struct sif_cq *cq;
    volatile struct sif_rq_sw *sw; /* kernel mapped sw descriptor entries */
    pthread_spinlock_t lock; /* Serializes operations on this queue */
};


/* TBD: For now just simple linear data structures for table block mappings
 *  - if necessary we might replace this with a more sophisticated
 *    (and dynamic) container
 */
#define SIF_MAX_BLOCKS (8*1024)

struct sif_desc_cache {
    enum desc_type type;
    struct sif_table ba;
    struct sif_map d[SIF_MAX_BLOCKS];
};

struct sif_context {
    struct verbs_context  verbs_ctx;
    enum sif_vendor_flags default_flags;
    enum sif_vendor_flags kernel_flags; /* Vendor flags set by the kernel */
    enum sif_mem_type mem_type;
    uint64_t map_length; /* Used by gva_type SIFGT_ZERO - indicates (optional) psif map length */
    uint64_t phys_length; /* Used by gva_type SIFGT_ZERO - indicates valid memory length */
    enum sif_proxy_type proxy; /* Used with proxy_mode to tell which EPS-A to proxy to */
    pthread_mutex_t lock; /* Protects the descriptor caches */
    struct sif_desc_cache dc[SIFD_MAX];
    struct sif_qp **qp_ref[SIF_MAX_BLOCKS];
    struct sif_mr **bypass_mr_ref;
    int bypass_cnt; /* Number of bypass MRs initiated - to avoid slow lookup unless bypass */
    uint64_t debug_mask;
    uint64_t opt_disable_mask;
    void (* fast_copy)(void *dst, const void *src, int len);
};

struct sif_pd {
    struct ibv_pd   ibv_pd;
    pthread_spinlock_t lock; /* Serializes access to the collect buffer */
    struct sif_map  cb;
#if HAVE_SHARED_PD
    uint8_t shpd; /* Used by sif_free_pd to discriminate between regular pd and shared pd */
#endif
};

enum sif_cq_type {
    SIF_CQ_NONE,
    SIF_CQ_REGULAR,	/* Persistent for the life of CQ */
    SIF_CQ_XRC
};

struct sif_cq {
    struct ibv_cq   ibv_cq;
    uint32_t        index;
    struct sif_map  m;
    uint32_t entries;
    uint32_t extent;
    uint32_t mask;     /* entries - 1 for modulo using & */
    uint32_t high_watermark;
    struct sif_rq *xsrq; /* WA #3796, for #3521, use a dedicated CQ for each XRCSRQ */
    enum sif_cq_type cq_type; /* WA #3796, for #3521 */
    volatile struct sif_cq_sw *sw; /* kernel mapped sw descriptor entries */
    pthread_spinlock_t lock; /* Serializes operations on this queue */
};

struct sif_qp {
    struct verbs_qp   verbs_qp;
    uint32_t qp_idx;
    uint32_t magic;
    uint8_t  qosl;
    uint8_t  tsl;
    uint8_t  force_sq_mode;  /* Set if forced SQ mode is set on this QP */
    uint16_t eps_tag;  /* Value to use for the eps_tag field (proxy_qp) */
    uint16_t remote_lid;  /* Remote lid for RC */
    uint32_t max_inline_data;
    uint32_t remote_qp;
    struct sif_sq sq;  /* kernel mapped queue spaces */
    struct sif_rq *rq;
};

struct sif_ah {
    struct ibv_ah   ibv_ah;
    int index;
    uint16_t dlid;
    uint8_t sl;
};

struct sif_mr {
    struct ibv_mr   ibv_mr;
    off_t uv2dma;
    enum sif_mem_type mem_type;
};

static inline uint16_t abi_major_ver(struct sif_device *ctx)
{
    return (ctx->abi_version >> 8) & 0xFF;
}

static inline uint16_t abi_minor_ver(struct sif_device *ctx)
{
    return ctx->abi_version & 0xFF;
}


static inline uint16_t abi_ver(uint16_t major, uint16_t minor)
{
    return ((major & 0xff) << 8) | (minor & 0xff);
}

struct sif_device *to_sdev(struct ibv_device *ibdev);

static inline struct sif_context *sif_get_context(struct ibv_context *ibctx)
{
    return container_of(ibctx, struct sif_context, verbs_ctx.context);
}

#define to_sctx(x) sif_get_context(x)

/* cast inlines */
#define to_sxxx(xxx, type)                      \
    ((struct sif_##type *)                      \
     ((void *) ib##xxx - offsetof(struct sif_##type, ibv_##xxx)))


static inline struct sif_pd *to_spd(struct ibv_pd *ibpd)
{
    return to_sxxx(pd, pd);
}

static inline struct sif_cq *to_scq(struct ibv_cq *ibcq)
{
    return to_sxxx(cq, cq);
}

static inline struct sif_mr *to_smr(struct ibv_mr *ibmr)
{
    return to_sxxx(mr, mr);
}

static inline struct sif_qp *to_sqp(struct ibv_qp *ibqp)
{
    return container_of(ibqp, struct sif_qp, verbs_qp.qp);
}

static inline struct sif_rq *to_ssrq(struct ibv_srq *ibsrq)
{
    return container_of(ibsrq, struct sif_rq, verbs_srq.srq);
}

static inline struct sif_ah *to_sah(struct ibv_ah *ibah)
{
    return to_sxxx(ah, ah);
}

static inline struct sif_qp *sq_to_qp(struct sif_sq *sq)
{
    return container_of(sq, struct sif_qp, sq);
}

static inline void sif_set_default_flags(struct ibv_context *ctx, enum sif_vendor_flags flags)
{
    to_sctx(ctx)->default_flags = flags;
}

static inline enum sif_vendor_flags sif_get_default_flags(struct ibv_context *ctx)
{
    return to_sctx(ctx)->default_flags;
}


#define sif_define_entry_funcs(type) \
static inline struct psif_##type##_entry \
        *get_##type##_entry(struct sif_##type *q, __u32 seq)\
{\
  return (struct psif_##type##_entry*) ((char*)q->m.base \
     + ((seq & q->mask) * q->extent));	\
}

sif_define_entry_funcs(cq)
sif_define_entry_funcs(rq)
sif_define_entry_funcs(sq)

int sif_query_device(struct ibv_context *context,
                  struct ibv_device_attr *attr);
int sif_query_port(struct ibv_context *context, uint8_t port,
                struct ibv_port_attr *attr);

struct ibv_pd *sif_alloc_pd(struct ibv_context *pd);
#if HAVE_SHARED_PD
struct ibv_shpd *sif_alloc_shpd(struct ibv_pd *pd, uint64_t share_key, struct ibv_shpd *shpd);
struct ibv_pd *sif_share_pd(struct ibv_context *context, struct ibv_shpd *shpd, uint64_t share_key);
#endif
int sif_free_pd(struct ibv_pd *pd);

struct ibv_mr *sif_reg_mr(struct ibv_pd *pd, void *addr,
                size_t length, int access);
struct ibv_mr *sif_rereg_mr(struct ibv_mr *mr, int flags,
                        struct ibv_pd *pd, void *addr, size_t length,
                        int access);
int sif_dereg_mr(struct ibv_mr *mr);
struct ibv_mw *sif_alloc_mw(struct ibv_pd *pd, enum ibv_mw_type type);
int sif_bind_mw(struct ibv_qp *qp, struct ibv_mw *mw, struct ibv_mw_bind *mw_bind);
int sif_dealloc_mw(struct ibv_mw *mw);
struct ibv_cq *sif_create_cq(struct ibv_context *context, int cqe,
                   struct ibv_comp_channel *channel, int comp_vector);
int sif_req_notify_cq(struct ibv_cq *cq, int solicited_only);
void sif_cq_event(struct ibv_cq *cq);
int sif_resize_cq(struct ibv_cq *cq, int cqe);
int sif_destroy_cq(struct ibv_cq *cq);
int sif_poll_cq(struct ibv_cq *cq, int ne, struct ibv_wc *wc);

struct ibv_qp *sif_create_qp(struct ibv_pd *pd,
                   struct ibv_qp_init_attr *attr);
struct ibv_qp *sif_create_qp_ex(struct ibv_context *context,
                   struct ibv_qp_init_attr_ex *qp_init_attr_ex);

int sif_query_qp(struct ibv_qp *qp, struct ibv_qp_attr *attr,
           int attr_mask, struct ibv_qp_init_attr *init_attr);
int sif_modify_qp(struct ibv_qp *qp, struct ibv_qp_attr *attr,
            int attr_mask);
int sif_destroy_qp(struct ibv_qp *qp);

int sif_post_send(struct ibv_qp *ibqp, struct ibv_send_wr *wr,
            struct ibv_send_wr **bad_wr);
int sif_post_recv(struct ibv_qp *ibqp, struct ibv_recv_wr *wr,
            struct ibv_recv_wr **bad_wr);

struct ibv_srq *sif_create_srq(struct ibv_pd *pd,
                 struct ibv_srq_init_attr *attr);
struct ibv_srq *sif_create_srq_ex(struct ibv_context *context,
                 struct ibv_srq_init_attr_ex *srq_init_attr_ex);

int sif_modify_srq(struct ibv_srq *srq, struct ibv_srq_attr *attr,
             int attr_mask);
int sif_query_srq(struct ibv_srq *srq, struct ibv_srq_attr *attr);
int sif_destroy_srq(struct ibv_srq *srq);
int sif_post_srq_recv(struct ibv_srq *srq, struct ibv_recv_wr *wr,
            struct ibv_recv_wr **bad_wr);

struct ibv_ah *sif_create_ah(struct ibv_pd *pd, struct ibv_ah_attr *attr);
int sif_destroy_ah(struct ibv_ah *ah);
int sif_attach_mcast(struct ibv_qp *qp, const union ibv_gid *gid,
                uint16_t lid);
int sif_detach_mcast(struct ibv_qp *qp, const union ibv_gid *gid,
                uint16_t lid);
void sif_async_event(struct ibv_async_event *event);

#if HAVE_DECL_IBV_QPT_XRC
void *sif_get_legacy_xrc(struct ibv_srq *srq);
void sif_set_legacy_xrc(struct ibv_srq *srq, void *legacy_xrc_srq);
#endif

#if defined(__cplusplus)
}
#endif
#endif  /* _SIF_H */
